import s3fs
import geopandas as gpd
import os
import tempfile
import fiona
import rasterio
from pathlib import Path
import numpy as np
from rasterio.mask import mask

fs = s3fs.S3FileSystem(anon=True)
bucket_name = "sdmlab"

# FINDING THE INTERSECTED HUC8 AND RETURNING GEOMETRY IN WGS84
def HUC8_inS3(fs, bucket, prefix="HUC8_boundaries/"):
    files = fs.ls(f"{bucket}/{prefix}")
    gpkg_key = next((f for f in files if f.endswith('.gpkg')), None)

    if gpkg_key is None:
        raise FileNotFoundError(f"No .gpkg file found in s3://{bucket}/{prefix}")

    # Download and read GeoPackage into a GeoDataFrame
    with fs.open(gpkg_key, 'rb') as s3file:
        with tempfile.NamedTemporaryFile(suffix=".gpkg", delete=False) as tmp_file:
            tmp_file.write(s3file.read())
            tmp_path = tmp_file.name
    
    gdf = gpd.read_file(tmp_path)
    return gdf

# WRAPPING ALL FUNCTIONS
def getHUC8BoundaryByID(huc_id):
    huc8_gdf = HUC8_inS3(fs, bucket_name)
    if huc8_gdf.crs != "EPSG:4326":
        huc8_gdf = huc8_gdf.to_crs("EPSG:4326")
    selected = huc8_gdf[huc8_gdf['HUC8'] == huc_id]
    if selected.empty:
        raise ValueError(f"HUC8 ID {huc_id} not found in the boundary file.")
    
    return selected.geometry


#PWB of CONUS rivers
def PWB_inS3(fs, bucket, prefix="PWB/"):
    tmp_dir = tempfile.mkdtemp()
    files = fs.ls(f"{bucket}/{prefix}")

    # Filter out relevant shapefile components
    for file_key in files:
        file_name = os.path.basename(file_key)
        if file_name.endswith(('.shp', '.shx', '.dbf', '.prj', '.cpg')):
            with fs.open(file_key, 'rb') as s3file:
                local_path = os.path.join(tmp_dir, file_name)
                with open(local_path, 'wb') as local_file:
                    local_file.write(s3file.read())

    # Ensure we got a .shp file
    shp_files = [f for f in os.listdir(tmp_dir) if f.endswith(".shp")]
    if not shp_files:
        raise ValueError("No .shp file found after download.")

    return os.path.join(tmp_dir, shp_files[0])

#GET FORCINGS
def get_forcings(huc_id, downloadforcings=True):
    s3_prefix = f"SM_dataset/HUCIDs_forcings/HUC{huc_id}/"  
    local_folder = Path(f"HUC{huc_id}_forcings")
    local_folder.mkdir(exist_ok=True)

    if downloadforcings:
        fs = s3fs.S3FileSystem(anon=True)
        bucket_name = "sdmlab"
        s3_files = fs.ls(f"{bucket_name}/{s3_prefix}")

        if not s3_files:
            raise FileNotFoundError(f"No files found at s3://{bucket_name}/{s3_prefix}")

        # Download all files
        for s3_path in s3_files:
            relative_path = Path(s3_path.replace(f"{bucket_name}/{s3_prefix}", ""))
            local_path = local_folder / relative_path
            local_path.parent.mkdir(parents=True, exist_ok=True)
            fs.get(s3_path, str(local_path))
    else:
        print(f"Skipping download; using existing benchmark data for HUC{huc_id}")

def get_population_GRID(boundary_gdf, fs = fs, bucket=bucket_name, prefix="SM_dataset/gridded_population/"):
    files = fs.ls(f"{bucket}/{prefix}")
    tif_key = next((f for f in files if f.endswith('.tif')), None)

    if tif_key is None:
        raise FileNotFoundError(f"No .tif file found in s3://{bucket}/{prefix}")

    with fs.open(tif_key, 'rb') as s3file:
        with tempfile.NamedTemporaryFile(suffix=".tif", delete=False) as tmp_file:
            tmp_file.write(s3file.read())
            tmp_tif_path = tmp_file.name

    # Clip gridded population raster with geometry
    with rasterio.open(tmp_tif_path) as src:
        geoms = [geom.__geo_interface__ for geom in boundary_gdf.geometry]
        out_image, out_transform = mask(src, geoms, crop=True)
        out_meta = src.meta.copy()

    out_meta.update({
        "height": out_image.shape[1],
        "width": out_image.shape[2],
        "transform": out_transform
    })
    return out_image, out_meta



import os
import rasterio

#INITIALIZE IN HUC EVENT DICT
def initialize_huc_event(huc_id, event_times):
    if isinstance(event_times, str):
        event_times = [event_times]
    return {huc_id: event_times}

#Create a folder and make it a working directory
def createCWD(folder_name):
    os.makedirs(folder_name, exist_ok=True)
    os.chdir(folder_name)
    new_path = os.getcwd()
    return new_path

def compress_tif_lzw(tif_path):
    # Read original file
    with rasterio.open(tif_path) as src:
        profile = src.profile.copy()
        data = src.read()
    profile.update(compress='lzw')

    with rasterio.open(tif_path, 'w', **profile) as dst:
        dst.write(data)





import os
import sys
import glob
import shutil
import subprocess
from dotenv import load_dotenv

from .datadownload import setup_directories


def runfim(code_dir, output_dir, HUC_code, data_dir, depth=False):
    original_dir = os.getcwd()
    try:
        tools_path = os.path.join(code_dir, "tools")
        src_path = os.path.join(code_dir, "src")
        os.chdir(tools_path)
        dotenv_path = os.path.join(code_dir, ".env")
        load_dotenv(dotenv_path)
        sys.path.append(src_path)
        sys.path.append(code_dir)

        HUC_code = str(HUC_code)
        HUC_dir = os.path.join(output_dir, f"flood_{HUC_code}")
        csv_path = data_dir

        discharge_basename = os.path.basename(data_dir).split(".")[0]
        inundation_dir = os.path.join(HUC_dir, f"{HUC_code}_inundation")
        temp_dir = os.path.join(inundation_dir, "temp")

        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)

        inundation_file = os.path.join(temp_dir, f"{discharge_basename}_inundation.tif")
        Command = [
            sys.executable,
            "inundate_mosaic_wrapper.py",
            "-y",
            HUC_dir,
            "-u",
            HUC_code,
            "-f",
            csv_path,
            "-i",
            inundation_file,
        ]

        if depth:
            depth_file = os.path.join(temp_dir, f"{discharge_basename}_depth.tif")
            Command += ["-d", depth_file]
        else:
            depth_file = None

        env = os.environ.copy()
        env["PYTHONPATH"] = f"{src_path}{os.pathsep}{code_dir}"

        result = subprocess.run(
            Command,
            cwd=tools_path,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        print(result.stdout.decode())
        if result.stderr:
            print(result.stderr.decode())

        if result.returncode == 0:
            print(f"Inundation mapping for {HUC_code} completed successfully.")

            if os.path.exists(inundation_file):
                dest_file = os.path.join(
                    inundation_dir, os.path.basename(inundation_file)
                )
                os.makedirs(inundation_dir, exist_ok=True)
                try:
                    os.replace(inundation_file, dest_file)
                except Exception:
                    if os.path.exists(dest_file):
                        os.remove(dest_file)
                    shutil.move(inundation_file, dest_file)

            if depth and depth_file and os.path.exists(depth_file):
                dest_depth = os.path.join(inundation_dir, os.path.basename(depth_file))
                try:
                    os.replace(depth_file, dest_depth)
                except Exception:
                    if os.path.exists(dest_depth):
                        os.remove(dest_depth)
                    shutil.move(depth_file, dest_depth)

            if os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)
        else:
            print(f"Failed to complete inundation mapping for {HUC_code}.")

    finally:
        os.chdir(original_dir)


def runOWPHANDFIM(huc, depth=False, version=None):
    code_dir, data_dir, output_dir = setup_directories()

    discharge = glob.glob(os.path.join(data_dir, f"*{huc}*.csv"))
    for file in discharge:
        runfim(code_dir, output_dir, huc, file, depth=depth)



import fimserve as fm
import pandas as pd
import time

huc_dir = pd.read_csv("../HUCs.csv")
SubSetHUC8 = huc_dir[:2].copy()

# To record the time
download_time = []
for i in SubSetHUC8["HUC8"]:
    start_time = time.time()
    fm.DownloadHUC8(i)
    end_time = time.time()
    download_time.append(end_time - start_time)

# Download times as a new column
SubSetHUC8["Downloading raster time"] = download_time
huc_dir.update(SubSetHUC8)
huc_dir.to_csv("../HUCs_updated.csv", index=False)


# Download the data
# fm.DownloadHUC8(huc)

# If user wants to update the streamorder

# stream_order = [6, 5]
# fm.DownloadHUC8(huc, stream_order)



import fimserve as fm

huc = "03020202"

start_date = "2020-01-01"
end_date = "2020-01-03"

# fm.getNWMretrospectivedata(huc, start_date, end_date)

# to work with only huc or for getting a retrospective streamflow for the evaluation or something
value_times = ["2020-01-01 00:00:00"]
# fm.getNWMretrospectivedata(huc, start_date, end_date, value_times)


# For the multiple watersheds with multiple events at the same time
huc_event_dict = {
    "03020202": ["2016-10-08 15:00:00", "2016-10-08"],
    "12060202": ["2016-10-09 15:00:00", "2016-10-09 16:00:00", "2016-10-09 17:00:00"],
}
# fm.getNWMretrospectivedata(huc_event_dict=huc_event_dict)
fm.getUSGSsitedata(huc, start_date, end_date)
